""" Contains perceptron learning methods for structured prediction. """


from __future__ import division

import logging
import array

import numpy

cimport numpy
cimport cython


__author__ = 'smartschat'


cdef class Perceptron:
    cdef int n_iter, random_seed
    cdef double cost_scaling
    cdef dict priors, weights
    """ Provide a latent structured perceptron.

    This implementation provides a latent structured perceptron with
    cost-augmented inference and parameter averaging for graphs encoding
    coreference decisions.

    Attributes:
        n_iter (int): The number of epochs for training. Defaults to 5.
        seed (int): The random seed for shuffling the data. Defaults to 23.
        cost_scaling (int): The parameter for scaling the cost function during
            cost-augmented inference. Defaults to 1.
        priors (dict(str, float)): A mapping of graph labels to priors for
            these labels.
        weights (dict(str, array)): A mapping of labels to weight
              vectors. For each label ``l``, ``weights[l]`` contains weights
              for each feature seen during training (for representing the
              features we employ *feature hashing*). If the graphs employed are
              not labeled, ``l`` is set to "+".
    """

    def __init__(self,
                 n_iter=5,
                 seed=23,
                 cost_scaling=1,
                 priors=None,
                 weights=None):
        """
        Initialize the perceptron.

        Args:
            n_iter (int): The number of epochs for training. Defaults to 5.
            seed (int): The random seed for shuffling the data. Defaults to 23.
            cost_scaling (int): The parameter for scaling the cost function
                during cost-augmented inference. Defaults to 1.
            label (list(str)): A list of labels used in the graphs. If
                ``None``, defaults to ``["+"]``.
            priors (dict(str, float)): A mapping of graph labels to priors
                for these labels. If ``None`` defaults to an empty mapping.
            weights (dict(str, array)): A mapping of graph labels to
                arrays. For each label ``l``, ``weights[l]`` contains
                weights for each feature seen during training (for representing
                the features we employ *feature hashing*). If the graphs
                employed are not labeled, ``l`` is set to "+".
                If ``None`` defaults to a mapping of "+" to an array only
                containing 0s.
        """
        cdef double[:] weights_for_label

        self.n_iter = n_iter
        self.random_seed = seed
        self.cost_scaling = cost_scaling

        labels = self.get_labels()

        if not priors:
            self.priors = {}
            for label in labels:
                self.priors[label] = 0.0
        else:
            self.priors = priors

        self.weights = {}

        try:
            if weights is None:
                for label in labels:
                    weights_for_label = array.array("d",
                                                        (0.0 for k in range(2**24)))
                    self.weights[label] = weights_for_label
            else:
                for label in weights:
                    weights_for_label = weights[label]
                    self.weights[label] = weights_for_label
        # in python 2, array.array does not support the buffer interface
        except TypeError:
            if weights is None:
                for label in labels:
                    weights_for_label = numpy.zeros(2**24, dtype=float)
                    self.weights[label] = weights_for_label
            else:
                for label in weights:
                    weights_for_label = numpy.array(weights[label])
                    self.weights[label] = weights_for_label

    def fit(self, substructures, arc_information):
        """Learn weights from data.

        Besides returning the learned model, also
        set the corresponding attributes ``self.priors``and ``self.weights``.

        Args:
            substructures (list(list((Mention, Mention)))): The search space
                for the substructures, defined by a nested list. The ith list
                contains the search space for the ith substructure.
            arc_information (dict((Mention, Mention), (array, int, bool)): A
                mapping of arcs (= mention pairs) to information about these
                arcs. The information consists of the features (represented as
                an int array via feature hashing), the costs for the arc, and
                whether predicting the arc to be coreferent is consistent with
                the gold annotation).

        Returns:
            A tuple describing the learned model, consisting of

                - **priors** (*dict(str, float)*): A mapping of graph labels to
                  priors for these labels.
                - **weights** (*dict(str, array)*): A mapping of graph
                  labels to arrays. For each label ``l``, ``weights[l]``
                  contains weights for each feature seen during training
                  (for representing the features we employ *feature hashing*).
                  If the graphs employed are not labeled, ``l`` is set to "+".
        """

        cdef double[:] cached_weights_for_label

        indices = list(range(0, len(substructures)))
        numpy.random.seed(self.random_seed)

        cached_priors = {}
        cached_weights = {}

        for label in self.priors:
            cached_priors[label] = 0.0

            try:
                cached_weights_for_label = array.array("d",
                                                    (0.0 for k in range(2**24)))
            except TypeError:
                cached_weights_for_label = numpy.zeros(2**24, dtype=float)

            cached_weights[label] = cached_weights_for_label

        counter = 0

        for epoch in range(1, self.n_iter+1):
            logging.info("Started epoch " + str(epoch))
            numpy.random.shuffle(indices)

            incorrect = 0

            for i in indices:
                substructure = substructures[i]

                (arcs,
                 arcs_labels,
                 arcs_scores,
                 cons_arcs,
                 cons_labels,
                 cons_scores,
                 is_consistent) = self.argmax(substructure,
                                              arc_information)

                if not is_consistent:
                    self.__update(cons_arcs,
                                  arcs,
                                  cons_labels,
                                  arcs_labels,
                                  arc_information,
                                  counter,
                                  cached_priors,
                                  cached_weights)
                    incorrect += 1

                counter += 1

            logging.info("Finished epoch " + str(epoch))
            logging.info("\tIncorrect predictions: " + str(incorrect) + "/" +
                         str(len(indices)))

        # averaging
        for label in self.priors:
            self.priors[label] -= (1/counter)*cached_priors[label]
            self._average_weights(self.weights[label], cached_weights[label],
                                  1.0*counter)

    def predict(self, substructures, arc_information):
        """
        Predict coreference information according to a learned model.

        Args:
            substructures (list(list((Mention, Mention)))): The search space
                for the substructures, defined by a nested list. The ith list
                contains the search space for the ith substructure.
            arc_information (dict((Mention, Mention), (array, int, bool)): A
                mapping of arcs (= mention pairs) to information about these
                arcs. The information consists of the features (represented as
                an int array via feature hashing), the costs for the arc, and
                whether predicting the arc to be coreferent is consistent with
                the gold annotation).
        Returns:
            Three nested lists describing the output. In particular, these
            lists are:

                - arcs (list(list(Mention, Mention))): The nested list of
                  predicted arcs. The ith list contains predictions for the
                  ith substructure.
                - labels (list(list(str))): Labels of the predicted arcs.
                - arcs (list(list(float))): Scores for the predicted arcs.
        """
        arcs = []
        labels = []
        scores = []

        for substructure in substructures:
            (substructure_arcs, substructure_arcs_labels,
             substructure_arcs_scores, _, _, _, _) = self.argmax(
                substructure, arc_information)

            arcs.append(substructure_arcs)
            labels.append(substructure_arcs_labels)
            scores.append(substructure_arcs_scores)

        return arcs, labels, scores

    def argmax(self, substructure, arc_information):
        """ Decoder for coreference resolution.

        Compute highest-scoring substructure and highest-scoring constrained
        substructure consistent with the gold annotation. To implement
        coreference resolution approaches, inherit this class and implement
        this function.

        Args:
            substructure (list((Mention, Mention))): The list of mention pairs
                which define the search space for one substructure.
            arc_information (dict((Mention, Mention), (array, int, bool)): A
                mapping of arcs (= mention pairs) to information about these
                arcs. The information consists of the features (represented as
                an int array via feature hashing), the costs for the arc, and
                whether predicting the arc to be coreferent is consistent with
                the gold annotation).

        Returns:
            A 6-tuple describing the highest-scoring substructure and the
            highest-scoring substructure consistent with the gold information.
            The tuple consists of:

                - **best_arcs** (*list((Mention, Mention))*): the list of arcs
                   in the highest-scoring substructure,
                - **best_labels** (*list(str)*): the list of labels of the
                  arcs in the highest-scoring substructure,
                - **best_scores** (*list(float)*): the scores of the arcs in
                  the highest-scoring substructure,
                - **best_cons_arcs** (*list((Mention, Mention))*): the list of
                  arcs in the highest-scoring constrained substructure
                  consistent with the gold information,
                - **best_cons_labels** (*list(str)*): the list of labels of the
                  arcs in the highest-scoring constrained substructure
                  consistent with the gold information,
                - **best_cons_scores** (*list(float)*): the scores of the arcs
                  in the highest-scoring constrained substructure consistent
                  with the gold information,
                - **is_consistent** (*bool*): whether the highest-scoring
                  substructure is consistent with the gold information.
        """
        raise NotImplementedError()

    def __update(self, good_arcs, bad_arcs, good_labels, bad_labels,
                 arc_information, counter, cached_priors, cached_weights):

        if good_labels or bad_labels:
            for arc, label in zip(good_arcs, good_labels):
                self._update_cython(self.weights[label],
                               cached_weights[label],
                               arc_information[arc][0],
                               1.0,
                               1.0*counter)

                self.priors[label] += 1
                cached_priors[label] += counter

            for arc, label in zip(bad_arcs, bad_labels):
                self._update_cython(self.weights[label],
                               cached_weights[label],
                               arc_information[arc][0],
                               -1.0,
                               -1.0*counter)

                self.priors[label] -= 1
                cached_priors[label] -= counter
        else:
            for arc in good_arcs:
                self._update_cython(self.weights["+"],
                               cached_weights["+"],
                               arc_information[arc][0],
                               1.0,
                               1.0*counter)

            for arc in bad_arcs:
                self._update_cython(self.weights["+"],
                               cached_weights["+"],
                               arc_information[arc][0],
                               -1.0,
                               -1.0*counter)

    def get_labels(self):
        return ["+"]

    def get_coref_labels(self):
        return ["+"]

    def find_best_arcs(self, arcs, arc_information, label="+"):
        max_val = float("-inf")
        best = None

        max_cons = float("-inf")
        best_cons = None

        best_is_consistent = False

        for arc in arcs:
            features, costs, consistent = arc_information[arc]

            score = self._cython_score_arc(self.priors[label],
                                      self.weights[label],
                                      self.cost_scaling,
                                      costs,
                                      features)

            if score > max_val:
                best = arc
                max_val = score
                best_is_consistent = consistent

            if score > max_cons and consistent:
                best_cons = arc
                max_cons = score

        return best, max_val, best_cons, max_cons, best_is_consistent

    def score_arc(self, features, costs, label="+"):
        """ Score an arc (described by features) according to priors, weights
        and costs.

        Args:
            features (numpy.array): An array containing integer features.
            costs (int): The costs of predicting the arc described by
                ``features``.
            label (str): The label of the arc. Defaults to "+".

        Returns:
            float: The sum of all weights for the features, plus the scaled
                costs for predicting the arc, plus the prior for the label.
        """

        return self._cython_score_arc(
            self.priors[label],
            self.weights[label],
            self.cost_scaling,
            costs,
            features
        )

    def get_model(self):
        array_weights = {}

        for label in self.weights:
            array_weights[label] = array.array("d", self.weights[label])

        return self.priors, array_weights


    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef double _cython_score_arc(self,
                                  double prior,
                                  double[:] weights,
                                  double cost_scaling,
                                  double costs,
                                  numpy.uint32_t[:] features) nogil:

        cdef double score = 0.0
        cdef int index = 0

        score += prior
        score += cost_scaling * costs

        for index in range(features.shape[0]):
            score += weights[features[index]]

        return score

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef void _update_cython(self,
                             double[:] weights,
                             double[:] cached_weights,
                             numpy.uint32_t[:] features,
                             double update_val_for_weights,
                             double update_val_for_cached_weights):
        cdef int index

        for index in range(features.shape[0]):
            weights[features[index]] += update_val_for_weights
            cached_weights[features[index]] += update_val_for_cached_weights

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef void _average_weights(self, double[:] weights, double[:] cached_weights,
                          double counter):
        cdef int index

        for index in range(weights.shape[0]):
            weights[index] -= cached_weights[index]/counter


""" Extract errors made by systems w.r.t. a reference corpus. """


from cort.analysis import data_structures


__author__ = 'smartschat'


class ErrorExtractor:
    """ Extract, manage and store recall and precision errors.

    Error extraction for recall errors works as follows:

    Go through each document. For each reference entity e in the document,
    construct an entity graph g_e for e and compute a partition of g_e by the
    system entity graphs. Then compute a spanning tree t_e of g_e and take
    every edge in t_e that does not appear in the partition as an error.

    For computing precision errors, switch the roles of reference and system
    entities.

    Attributes:
        reference_corpus (Corpus): The reference corpus with the gold
            information concerning the coreference relation.
        recall_spanning_tree_algorithm (function): A function mapping an
            entity graph and one its partitions to a list of mentions pairs,
            which represent a spanning tree of the entity graph. This
            function is used to compute recall errors.
        precision_spanning_tree_algorithm (function): Same as above, but for
            precision errors.
        errors (dict): A mapping of error descriptions to sets containing the
            respective errors.
    """
    def __init__(self,
                 reference_corpus,
                 recall_spanning_tree_algorithm,
                 precision_spanning_tree_algorithm,
                 ):
        """ Initialize the error analysis.

        Args:
            reference_corpus (Corpus): The reference corpus with the gold
                information concerning the coreference relation.
            recall_spanning_tree_algorithm (function): A function mapping an
                entity graph and one its partitions to a list of mentions pairs,
                which represent a spanning tree of the entity graph. This
                function is used to compute recall errors.
            precision_spanning_tree_algorithm (function): Same as above, but for
                precision errors.
        """

        self.reference_corpus = reference_corpus
        self.recall_spanning_tree_algorithm = recall_spanning_tree_algorithm
        self.precision_spanning_tree_algorithm = \
            precision_spanning_tree_algorithm
        self.errors = {}
        self.corpora = {}

    def add_system(self, system_corpus, which_mentions="annotated"):
        """ Add a system to the error analysis.

        Error extraction for recall errors works as follows:

        Go through each document. For each reference entity e in the document,
        construct an entity graph g_e for e and compute a partition of g_e by
        the system entity graphs. Then compute a spanning tree t_e of g_e and
        take every edge in t_e that does not appear in the partition as an
        error.

        For computing precision errors, switch the roles of reference and system
        entities.

        Args:
            system_corpus (Corpus): A corpus obtained from system output.
            which_mentions (str): Either "annotated" or "extracted",
                defaults to "annotated". Specifies from which mentions in
                the system corpus coreference information should be
                obtained, either annotated mentions or system mentions.
        """
        if which_mentions not in ["annotated", "extracted"]:
            raise ValueError("which_mentions must be"
                             "either 'annotated' or 'extracted'.")

        recall_errors, precision_errors = self.__compute_errors(system_corpus,
                                                                which_mentions)

        self.errors[system_corpus.description] = {
            "recall_errors": {},
            "precision_errors": {}
        }

        self.errors[system_corpus.description]["recall_errors"]["all"] = \
            recall_errors
        self.errors[
            system_corpus.description]["precision_errors"]["all"] = \
            precision_errors

        self.corpora[system_corpus.description] = system_corpus

    def get_errors(self):
        """ Get errors for all systems managed by this ErrorAnalysis.

        The errors are stored via an ``StructuredCoreferenceAnalysis`
        which can be accessed like a dict.

        If a corpus with the description
        ``ranking``was added via ``self.add_system``,
        ``self.errors["ranking"]["recall_errors"]["all"]``is an ``EnhancedSet``
        containing all recall errors of the system. Errors of other systems
        ańd precision errors can be accessed analogously.

        Returns:
            StructuredCoreferenceAnalysis: A StructuredCoreferenceAnalysis
                containing the errors.
        """
        return data_structures.StructuredCoreferenceAnalysis(
            self.errors, corpora=self.corpora,
            reference=self.reference_corpus)

    def __compute_errors(self, system_corpus, which_mentions):
        gold_graphs = [data_structures.EntityGraph.from_mentions(
            doc.annotated_mentions, "annotated_set_id")
            for doc in self.reference_corpus.documents]

        if which_mentions == 'annotated':
            system_graphs = [data_structures.EntityGraph.from_mentions(
                doc.annotated_mentions, "annotated_set_id")
                for doc in system_corpus.documents]
        else:
            system_graphs = [data_structures.EntityGraph.from_mentions(
                doc.system_mentions, "set_id")
                for doc in system_corpus.documents]

        recall_errors = []
        precision_errors = []

        for doc_gold_graphs, doc_system_graphs in zip(gold_graphs,
                                                      system_graphs):
            recall_errors.extend(
                self.__compute_errors_for_doc(
                    doc_gold_graphs,
                    doc_system_graphs,
                    self.recall_spanning_tree_algorithm))
            precision_errors.extend(
                self.__compute_errors_for_doc(
                    doc_system_graphs,
                    doc_gold_graphs,
                    self.precision_spanning_tree_algorithm))

        return (data_structures.EnhancedSet(recall_errors),
                data_structures.EnhancedSet(precision_errors))

    @staticmethod
    def __compute_errors_for_doc(base_graphs,
                                 partitioning_graphs,
                                 spanning_tree_algorithm):
        errors = []

        for graph in base_graphs:
            errors.extend(
                ErrorExtractor.__compute_errors_for_graph(
                    graph, partitioning_graphs, spanning_tree_algorithm))

        return errors

    @staticmethod
    def __compute_errors_for_graph(graph,
                                   partitioning_graphs,
                                   spanning_tree_algorithm):
        partitioned_graph = graph.partition(partitioning_graphs)
        spanning_tree = spanning_tree_algorithm(graph, partitioned_graph)
        extra_pairs = [
            (anaphor, antecedent) for anaphor, antecedent in spanning_tree
            if anaphor not in partitioned_graph.edges or
            antecedent not in partitioned_graph.edges[anaphor]
        ]

        return [(anaphor, antecedent) for anaphor, antecedent in sorted(
            extra_pairs)]



from __future__ import print_function

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import subprocess
import tempfile
from .StanfordDependencies import (StanfordDependencies,
                                   JavaRuntimeVersionError)
from .CoNLL import Corpus

JAVA_CLASS_NAME = 'edu.stanford.nlp.trees.EnglishGrammaticalStructure'

class SubprocessBackend(StanfordDependencies):
    """Interface to Stanford Dependencies via subprocesses. This means
    that each call opens a pipe to Java. It has the advantage that it
    should work out of the box if you have Java but it is slower than
    other backends. As such, convert_trees() will be more efficient than
    convert_tree() for this backend."""
    def __init__(self, jar_filename=None, download_if_missing=False,
                 version=None, java_command='java'):
        """java_command is the path to a java binary."""
        StanfordDependencies.__init__(self, jar_filename, download_if_missing,
                                      version)
        self.java_command = java_command
    def convert_trees(self, ptb_trees, representation='basic',
                      include_punct=True, include_erased=False, debug=False):
        """Convert a list of Penn Treebank formatted trees (ptb_trees)
        into Stanford Dependencies. The dependencies are represented
        as a list of sentences, where each sentence is itself a list of
        Token objects.

        Currently supported representations are 'basic', 'collapsed',
        'CCprocessed', and 'collapsedTree' which behave the same as they
        in the CoreNLP command line tools.

        Setting debug=True will cause debugging information (including
        the java command run to be printed."""
        self._raise_on_bad_representation(representation)
        input_file = tempfile.NamedTemporaryFile(delete=False)
        try:
            for ptb_tree in ptb_trees:
                tree_with_line_break = ptb_tree + "\n"
                input_file.write(tree_with_line_break.encode("UTF-8"))
            input_file.flush()

            command = [self.java_command,
                       '-ea',
                       '-cp', self.jar_filename,
                       JAVA_CLASS_NAME,
                       '-' + representation,
                       '-treeFile', input_file.name]
            # if we're including erased, we want to include punctuation
            # since otherwise we won't know what SD considers punctuation
            if include_punct or include_erased:
                command.append('-keepPunct')
            if debug:
                print('Command:', ' '.join(command))
            sd_process = subprocess.Popen(command, stdout=subprocess.PIPE,
                                          stderr=subprocess.PIPE)
            return_code = sd_process.wait()
            stderr = sd_process.stderr.read()
            stdout = sd_process.stdout.read()
            self._raise_on_bad_exitcode(return_code, stderr, debug)
        finally:
            os.remove(input_file.name)

        sentences = Corpus.from_stanford_dependencies(stdout.splitlines(),
                                                      ptb_trees,
                                                      include_erased,
                                                      include_punct)

        assert len(sentences) == len(ptb_trees), \
            "Only got %d sentences from Stanford Dependencies when " \
            "given %d trees." % (len(sentences), len(ptb_trees))
        return sentences
    def convert_tree(self, ptb_tree, **kwargs):
        """Converts a single Penn Treebank format tree to Stanford
        Dependencies. See convert_trees for more details."""
        return self.convert_trees([ptb_tree], **kwargs)[0]

    @staticmethod
    def _raise_on_bad_exitcode(return_code, stderr, debug=False):
        if debug:
            print('Exit code:', return_code)
            if stderr.strip():
                print('stderr:', stderr)

        if return_code:
            if 'Unsupported major.minor version' in stderr:
                # Oracle Java error message
                raise JavaRuntimeVersionError()
            elif 'JVMCFRE003 bad major version' in stderr:
                # IBM Java error message
                raise JavaRuntimeVersionError()
            else:
                raise ValueError("Bad exit code from Stanford CoreNLP")



""" Contains features for coreference resolution."""


import re


from cort.core import spans


__author__ = 'smartschat'


def fine_type(mention):
    """ Compute fine-grained type of a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        str: The string 'fine_type=TYPE', where TYPE is one of 'NAM', 'DEF',
        'INDEF', 'DEM', 'VRB', 'i', 'you', 'he', 'she', 'it', 'we', 'they'
        and 'NONE'.
    """
    if mention.attributes["type"] == "NOM":
        mention_fine_type = mention.attributes["fine_type"]
    elif mention.attributes["type"] == "PRO":
        mention_fine_type = mention.attributes["citation_form"]
    else:
        mention_fine_type = mention.attributes["type"]

    if not mention_fine_type:
        mention_fine_type = "NONE"

    return "fine_type=" + mention_fine_type


def gender(mention):
    """ Compute gender of a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        str: The string 'gender=GENDER', where GENDER is one of 'MALE',
        'FEMALE', 'NEUTRAL', 'PLURAL' and 'UNKNOWN'.
    """
    return "gender=" + mention.attributes["gender"]


def number(mention):
    """ Compute number of a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        str: The string 'number=NUMBER', where NUMBER is one of 'SINGULAR',
        'PLURAL' and 'UNKNOWN'.
    """
    return "number=" + mention.attributes["number"]


def sem_class(mention):
    """ Compute semantic class of a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        str: The string 'sem_class=SEM_CLASS', where SEM_CLASS is one of
        'PERSON', 'OBJECT', 'NUMERIC' and 'UNKNOWN'.
    """
    return "sem_class=" + mention.attributes["semantic_class"]


def gr_func(mention):
    """ Compute grammatical function of a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        str: The string 'gr_func=GR_FUNC', where GR_FUNC is one of 'SUBJECT',
        'OBJECT' and 'OTHER'.
    """
    return "gr_func=" + mention.attributes["grammatical_function"]


def governor(mention):
    """ Compute grammatical function of a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        str: The string 'gr_func=GR_FUNC', where GR_FUNC is one of 'SUBJECT',
        'OBJECT' and 'OTHER'.
    """
    return "governor=" + mention.attributes["governor"].lower()


def deprel(mention):
    return "deprel=" + mention.attributes["deprel"]


def head(mention):
    """ Compute head of a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        str: The string 'head=HEAD', where HEAD is the (lowercased) head of
        the mention.
    """
    return "head=" + mention.attributes["head_as_lowercase_string"]


def head_ner(mention):
    """ Compute named entity tag of a mention's head.

    Args:
        mention (Mention): A mention.

    Returns:
        str: The string 'ner=NER', where NER is the named entity tag of the
        mention's head word. If the mention is not a named entity, NER is
        set to 'NONE'.
    """
    return "ner=" + mention.attributes["ner"][mention.attributes["head_index"]]


def length(mention):
    """ Compute length of a mention in tokens.

    Args:
        mention (Mention): A mention.

    Returns:
        str: The string 'length=LENGTH', where LENGTH is the length of the
        mention in tokens.
    """
    return "length=" + str(len(mention.attributes["tokens"]))


def first(mention):
    """ Compute the first token of a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        str: The string 'first=TOKEN', where TOKEN is the first token of
        the mention.
    """
    return "first=" + mention.attributes["tokens"][0].lower()


def last(mention):
    """ Compute the last token of a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        str: The string 'last=TOKEN', where TOKEN is the last token of
        the mention..
    """
    return "last=" + mention.attributes["tokens"][-1].lower()


def preceding_token(mention):
    """ Compute the token preceding a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        str: The string 'preceding=TOKEN', where TOKEN is the token
        preceding the mention. If no such token exists, set TOKEN to 'NONE'.
    """
    prec = mention.get_context(-1)

    if prec:
        return "preceding=" + prec[0].lower()
    else:
        return "preceding=NONE"


def next_token(mention):
    """ Compute the token following a mention.

    Args:
        mention (Mention): A mention.

    Returns:
        str: The string 'next=TOKEN', where TOKEN is the token
        following the mention. If no such token exists, set TOKEN to 'NONE'.
    """
    next_t = mention.get_context(1)

    if next_t:
        return "next=" + next_t[0].lower()
    else:
        return "next=NONE"


def ancestry(mention):
    return "ancestry=" + mention.attributes["ancestry"]


def exact_match(anaphor, antecedent):
    """ Compute whether the tokens of two mentions match exactly.

    Args:
        anaphor (Mention): A mention.
        antecedent (Mention): Another mention, preceding the anaphor.

    Returns:
        str: 'exact_match' if the tokens of anaphor and antecedent
        match exactly (ignoring case), None otherwise.
    """
    if (anaphor.attributes["tokens_as_lowercase_string"] ==
            antecedent.attributes["tokens_as_lowercase_string"]):
        return "exact_match"


def head_match(anaphor, antecedent):
    """ Compute whether the heads of two mentions match.

    Args:
        anaphor (Mention): A mention.
        antecedent (Mention): Another mention, preceding the anaphor.

    Returns:
        str: 'head_match' if the heads of anaphor and antecedent
        match (ignoring case), None otherwise.
    """
    if (anaphor.attributes["head_as_lowercase_string"] ==
            antecedent.attributes["head_as_lowercase_string"]):
        return "head_match"


def tokens_contained(anaphor, antecedent):
    ana_tokens = anaphor.attributes["tokens_as_lowercase_string"]
    ante_tokens = antecedent.attributes["tokens_as_lowercase_string"]

    if ana_tokens in ante_tokens or ante_tokens in ana_tokens:
        return "tokens_contained"


def head_contained(anaphor, antecedent):
    ana_head = anaphor.attributes["head_as_lowercase_string"]
    ante_head = antecedent.attributes["head_as_lowercase_string"]

    if ana_head in ante_head or ante_head in ana_head:
        return "head_contained"


def sentence_distance(anaphor, antecedent):
    """ Compute the sentence distance between two mentions (capped at 5).

    Args:
        anaphor (Mention): A mention.
        antecedent (Mention): Another mention, preceding the anaphor.

    Returns:
        str: 'sentence_distance=DIST', where DIST is one of '0', '1',
        '2', '3', '4' and '>=5'.
    """
    return "sentence_distance=" + __compute_sentence_distance(anaphor, antecedent)


def token_distance(anaphor, antecedent):
    """ Compute the token distance between two mentions (capped at 10).

    Args:
        anaphor (Mention): A mention.
        antecedent (Mention): Another mention, preceding the anaphor.

    Returns:
        str: 'sentence_distance=DIST', where DIST is one of '0', '1',
        '2', '3', '4' and '>=5'.
    """
    return "token_distance=" + __compute_token_distance(anaphor, antecedent)

def alias(anaphor, antecedent):
    """ Compute whether the mentions are aliases of each other.

    Args:
        anaphor (Mention): A mention.
        antecedent (Mention): Another mention, preceding the anaphor.

    Returns:
        str: 'alias' if anaphor and antecedent are in an alias
            relation. None otherwise.
    """
    if __are_alias(anaphor, antecedent):
        return "alias"


def same_speaker(anaphor, antecedent):
    """
    Compute whether the speakers  of two mentions are the same..

    Args:
        anaphor (Mention): A mention.
        antecedent (Mention): Another mention, preceding the anaphor.

    Returns:
        str: 'same speaker' if the mentions have the same speaker,
        None otherwise.
    """
    if anaphor.attributes["speaker"] == \
            antecedent.attributes["speaker"]:
        return "same_speaker"


def __compute_sentence_distance(anaphor, antecedent):
    dist = anaphor.attributes['sentence_id'] - antecedent.attributes[
        'sentence_id']
    if dist >= 5:
        return ">=5"
    else:
        return str(dist)


def __compute_token_distance(anaphor, antecedent):
    dist = anaphor.span.begin - antecedent.span.end
    if dist >= 10:
        return ">=10"
    else:
        return str(dist)


def embedding(anaphor, antecedent):
    """ Compute whether one mention embeds the other.

    Args:
        anaphor (Mention): A mention.
        antecedent (Mention): Another mention, preceding the anaphor.

    Returns:
        str: 'embedding' if one of the mentions embeds the other,
        None otherwise.
    """
    if (anaphor.span.embeds(antecedent.span) or
            antecedent.span.embeds(anaphor.span)):
        return "embedding"


def modifier(anaphor, antecedent):
    """ Compute modifier agreement.

    In particular, compute whether the anaphor has modifiers that do not appear
    in the antecedent (ignoring demonstratives and prepositions).

    Args:
        anaphor (Mention): A mention.
        antecedent (Mention): Another mention, preceding the anaphor.

    Returns:
        str: 'modifier' if the anaphor has modifiers that do not
        appear in the antecedent. None otherwise.
    """
    if not __get_modifier(anaphor).issubset(__get_modifier(antecedent)):
        return "modifier"


def __get_modifier(mention):
    head_span_in_mention = spans.Span(
        mention.attributes["head_span"].begin - mention.span.begin,
        mention.attributes["head_span"].end - mention.span.begin)

    modifiers = set()

    for index, (token, pos) in enumerate(
            zip(mention.attributes["tokens"], mention.attributes["pos"])):
        if (token.lower() not in ["the", "this", "that", "those", "these",
                                  "a", "an"]
            and pos not in ["POS", "IN"]
            and (index < head_span_in_mention.begin
                 or index > head_span_in_mention.end)):
            modifiers.add(token.lower())

    return modifiers


def __are_alias(anaphor, antecedent):
    if anaphor.attributes["type"] != "NAM" or antecedent.attributes["type"] \
            != "NAM":
        return False
    elif anaphor.attributes["head_as_lowercase_string"] == \
            antecedent.attributes["head_as_lowercase_string"]:
        return False
    else:
        anaphor_cleaned_tokens = anaphor.attributes["head"]
        antecedent_cleaned_tokens = antecedent.attributes["head"]

        category = __get_category_for_alias(
            anaphor.attributes["ner"][
                anaphor.attributes["head_index"]
            ], antecedent.attributes["ner"][
                antecedent.attributes["head_index"]]
        )

        if category == "PERSON":
            return __person_alias(anaphor_cleaned_tokens,
                                  antecedent_cleaned_tokens)
        elif category == "LOC":
            return __loc_alias(anaphor_cleaned_tokens,
                               antecedent_cleaned_tokens)
        elif category == "ORG":
            return __org_alias(anaphor_cleaned_tokens,
                               antecedent_cleaned_tokens)
        else:
            return False


def __get_category_for_alias(anaphor_ner, antecedent_ner):
    if anaphor_ner == "PERSON" and antecedent_ner == "PERSON":
        return "PERSON"
    elif re.match(r"LOC", anaphor_ner) and re.match(r"LOC", antecedent_ner):
        return "LOC"
    elif re.match(r"ORG", anaphor_ner) and re.match(r"(ORG)", antecedent_ner):
        return "ORG"


def __loc_alias(anaphor_cleaned_tokens, antecedent_cleaned_tokens):
    return __starts_with(anaphor_cleaned_tokens, antecedent_cleaned_tokens) or \
        __is_abbreviation(anaphor_cleaned_tokens, antecedent_cleaned_tokens)


def __org_alias(anaphor_cleaned_tokens, antecedent_cleaned_tokens):
    return __starts_with(anaphor_cleaned_tokens, antecedent_cleaned_tokens) or \
        __is_abbreviation(anaphor_cleaned_tokens, antecedent_cleaned_tokens)


def __person_alias(anaphor_cleaned_tokens, antecedent_cleaned_tokens):
    if len(anaphor_cleaned_tokens) == 1 or len(antecedent_cleaned_tokens) == 1:
        return anaphor_cleaned_tokens[0] == antecedent_cleaned_tokens[0] \
            or anaphor_cleaned_tokens[-1] == antecedent_cleaned_tokens[-1]
    elif (len(anaphor_cleaned_tokens) == 2 and anaphor_cleaned_tokens[
        0].lower() in ["mr", "ms", "mr.", "ms."]
        or len(antecedent_cleaned_tokens) == 2 and antecedent_cleaned_tokens[
            0].lower() in ["mr", "ms", "mr.", "ms."]):
        return anaphor_cleaned_tokens[-1] == antecedent_cleaned_tokens[-1]
    elif anaphor_cleaned_tokens[0] == antecedent_cleaned_tokens[0] and \
            anaphor_cleaned_tokens[-1] == antecedent_cleaned_tokens[-1]:
        return True
    elif len(anaphor_cleaned_tokens) > 1 and len(antecedent_cleaned_tokens) > 1:
        return anaphor_cleaned_tokens[-1] == antecedent_cleaned_tokens[-1] and \
            anaphor_cleaned_tokens[-2] == antecedent_cleaned_tokens[-2]

    return False


def __starts_with(anaphor_cleaned_tokens, antecedent_cleaned_tokens):
    for ana_token, ante_token in zip(anaphor_cleaned_tokens,
                                     antecedent_cleaned_tokens):
        if ana_token != ante_token:
            return False

    return True


def __is_abbreviation(anaphor_cleaned_tokens, antecedent_cleaned_tokens):
    if " ".join(anaphor_cleaned_tokens).replace(".", "") == " ".join(
            antecedent_cleaned_tokens).replace(".", ""):
        return True
    else:
        if len(anaphor_cleaned_tokens) > len(antecedent_cleaned_tokens):
            return " ".join(antecedent_cleaned_tokens) in set(
                __get_acronyms(anaphor_cleaned_tokens))
        else:
            return " ".join(anaphor_cleaned_tokens) in set(
                __get_acronyms(antecedent_cleaned_tokens))


def __get_acronyms(cleaned_tokens):
    company_designator = r'assoc|bros|co|coop|corp|devel|inc|llc|ltd\.?'
    tokens_without_designator = [token for token in cleaned_tokens if
                                 not re.match(company_designator,
                                              token.lower())]

    return " ".join(tokens_without_designator), \
           "".join([token[0] for token in tokens_without_designator if
                    token[0].isupper()]), \
           ".".join([token[0] for token in tokens_without_designator if
                     token[0].isupper()]) + "."



""" Extract instances and features from a corpus. """


import array
import multiprocessing
import sys


import mmh3
from pickle import PicklingError
import numpy


__author__ = 'martscsn'


# for python 2 multiprocessing
def unwrap_extract_doc(arg, **kwarg):
    return InstanceExtractor._extract_doc(*arg, **kwarg)


class InstanceExtractor:
    """ Extract instances and their corresponding features from a corpus.

    Attributes:
        extract_substructures (function: CoNLLDocument ->
             list(list((Mention,Mention)))): Function for extracting the search
             space for a coreference resolution approach. The ith list in the
             nested list contains the search space for the ith substructure.
             The search space is represented as a nested list of mention pairs,
             which are candidate arcs in the graph to predict.
        mention_features (list(function: Mention -> str)): A list of features
            for mentions.
        pairwise_features (list(function: (Mention, Mention) -> str)): A list
            of features for mention pairs.
        cost_function (function: (Mention, Mention) -> int): A function
            assigning costs to mention pairs.
    """
    def __init__(self,
                 extract_substructures,
                 mention_features,
                 pairwise_features,
                 cost_function):
        """ Initialize instance and feature extraction.

        Args:
            extract_substructures (function: CoNLLDocument ->
                list(list((Mention,Mention)))): Function for extracting the
                search space for a coreference resolution approach. The ith
                list in the nested list contains the search space for the ith
                substructure. The search space is represented as a nested list
                of mention pairs, which are candidate arcs in the graph to
                predict.
            mention_features (list(function: Mention -> str)): A list of
                features for mentions.
            pairwise_features (list(function: (Mention, Mention) -> str)): A
                list of features for mention pairs.
            cost_function (function: (Mention, Mention) -> int): A function
                assigning costs to mention pairs.
        """
        self.extract_substructures = extract_substructures
        self.mention_features = mention_features
        self.pairwise_features = pairwise_features
        self.cost_function = cost_function

    def extract(self, corpus):
        """ Extract instances and features from a corpus.

        Args:
            corpus (Corpus): The corpus to extract instances and features from.

        Returns:
            A 6-tuple which describes the extracted instances and their
            features. The individual components are:

                - **arcs** (*list((Mention, Mention))*): A list of coreference
                  arcs.
                - **features** (*numpy.array*): An array of features. The features
                  are represented as integers via feature hashing.
                - **substructures_mapping** (*numpy.array*): A list of indices
                  that encodes which arcs in the list of arcs correspond to the
                  same substructure. For example, if
                  ``substructres_mapping[3] = 20`` and
                  ``substructures_mapping[4] = 100``, then ``arcs[20:100]``
                  correspond to the same substructure.
                - **arcs_mapping** (*numpy.array*): A list of indices that encodes
                  which features in the list of features correspond to the
                  same arc (following the same principle as
                  ``substructures_mapping``).
                - **coreferent** (*numpy.array*): A boolean array of size
                  len(arcs) containing the information whether the two mentions
                  in the arc are coreferent.
                - **costs** (*numpy.array*): An array of size len(arcs)
                  containing the costs of wrongly predicting an arc.
        """

        all_substructures = []
        arc_information = {}

        id_to_doc_mapping = {}
        for doc in corpus:
            id_to_doc_mapping[doc.identifier] = doc

        slice_size = len(corpus.documents)

        for num in range(0, len(corpus.documents), slice_size):
            docs = corpus.documents[num:num+slice_size]

            pool = multiprocessing.Pool(maxtasksperchild=1)

            if sys.version_info[0] == 2:
                results = pool.map(unwrap_extract_doc,
                                   zip([self]*len(corpus.documents),
                                       corpus.documents))
            else:
                results = pool.map(self._extract_doc, docs)

            pool.close()
            pool.join()

            for result in results:
                doc_identifier, anaphors, antecedents, features, costs, \
                    consistency, feature_mapping, substructures_mapping = result

                doc = id_to_doc_mapping[doc_identifier]

                for i in range(0, len(substructures_mapping)-1):
                    struct = []
                    begin = substructures_mapping[i]
                    end = substructures_mapping[i+1]

                    for pair_index in range(begin, end):
                        arc = (doc.system_mentions[anaphors[pair_index]],
                               doc.system_mentions[antecedents[pair_index]])

                        struct.append(arc)

                        features_start = feature_mapping[pair_index]
                        features_end = feature_mapping[pair_index+1]

                        arc_information[arc] = \
                            (features[features_start:features_end],
                             costs[pair_index],
                             consistency[pair_index])

                    all_substructures.append(struct)

        # in python 2, array.array does not support the buffer interface
        if sys.version_info[0] == 2:
            for arc in arc_information:
                feats, cost, cons = arc_information[arc]
                arc_information[arc] = (numpy.array(feats,
                                                    dtype=numpy.uint32),
                                        cost,
                                        cons)

        return all_substructures, arc_information

    def _extract_doc(self, doc):
        cache = {}
        substructures = self.extract_substructures(doc)

        mentions_to_ids = {}

        for i, mention in enumerate(doc.system_mentions):
            mentions_to_ids[mention] = i

        anaphors = array.array('H')
        antecedents = array.array('H')
        costs = array.array('H')
        consistency = array.array('B')
        feature_mapping = array.array('I')
        substructures_mapping = array.array('I')
        features = array.array('I')

        feature_mapping.append(0)
        substructures_mapping.append(0)

        for struct in substructures:
            # skip empty
            if not struct:
                continue

            for arc in struct:
                anaphors.append(mentions_to_ids[arc[0]])
                antecedents.append(mentions_to_ids[arc[1]])
                costs.append(self.cost_function(arc))
                consistency.append(arc[0].decision_is_consistent(arc[1]))

                arc_features = self._extract_features(arc, cache)
                features.extend(arc_features)
                feature_mapping.append(feature_mapping[-1] + len(arc_features))

            substructures_mapping.append(substructures_mapping[-1] +
                                         len(struct))

        return (doc.identifier,
                anaphors,
                antecedents,
                features,
                costs,
                consistency,
                feature_mapping,
                substructures_mapping)

    def _extract_features(self, arc, cache):
        anaphor, antecedent = arc
        inst_feats = []

        if not antecedent.is_dummy():
            # mention features
            for mention in [anaphor, antecedent]:
                if mention not in cache:
                    cache[mention] = [feature(mention) for feature
                                      in self.mention_features]

            ana_features = cache[anaphor]
            ante_features = cache[antecedent]

            inst_feats += ["ana_" + feat for feat in ana_features]
            inst_feats += ["ante_" + feat for feat in ante_features]

            # concatenated features
            inst_feats += ["ana_" + feat_ana + "^ante_" + feat_ante
                           for feat_ana, feat_ante in
                           zip(ana_features, ante_features)]

            # pairwise features
            inst_feats += [feature(anaphor, antecedent) for feature
                           in self.pairwise_features
                           if feature(anaphor, antecedent)]

            # feature combinations
            fine_type_indices = {len(self.mention_features)*i for i
                                 in [0, 1, 2]}

            inst_feats += [
                inst_feats[i] + "^" + word for i in fine_type_indices
                for j, word in enumerate(inst_feats)
                if j not in fine_type_indices
            ]

        # to hash
        all_feats = array.array('I', [mmh3.hash(word.encode("utf-8")) & 2**24-1
                                      for word in inst_feats])

        return all_feats



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple
import re

# picks out (tag, word) from Penn Treebank-style trees
ptb_tags_and_words_re = re.compile(r'\(\s*([^\s()]+)\s+([^\s()]+)\s*\)')

# picks out (deprel, gov, govindex, dep, depindex) from Stanford
# Dependencies text (e.g., "nsubj(word-1, otherword-2)")
deps_re = re.compile(r'^\s*([^\s()]+)\(([^\s()]+)-(\d+),\s+'
                     r'([^\s()]+)-(\d+)\)\s*$',
                     re.M)

# CoNLL-X field names
FIELD_NAMES = ('index', 'form', 'lemma', 'cpos', 'pos', 'feats', 'head',
               'deprel', 'phead', 'pdeprel')

class Token(namedtuple('Token', FIELD_NAMES)):
    """CoNLL-X style dependency token. Fields include:
    - form (the word form)
    - lemma (the word's base form or lemma) -- empty for SubprocessBackend
    - pos (part of speech tag)
    - index (index of the token in the sentence)
    - head (index of the head of this token), and
    - deprel (the dependency relation between this token and its head)

    There are other fields but they typically won't be populated by
    StanfordDependencies.

    See http://ilk.uvt.nl/conll/#dataformat for a complete description."""
    def __repr__(self):
        """Represent this Token as Python code. Note that the resulting
        representation may not be a valid Python call since this skips
        fields with empty values."""
        # slightly different from the official tuple __repr__ in that
        # we skip any fields with None as their value
        items = [(field, getattr(self, field, None)) for field in FIELD_NAMES]
        fields = ['%s=%r' % (k, v) for k, v in items if v is not None]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(fields))
    def as_conll(self):
        """Represent this Token as a line as a string in CoNLL-X format."""
        def get(field):
            value = getattr(self, field)
            if value is None:
                value = '_'
            elif field == 'feats':
                value = '|'.join(value)
            return str(value)
        return '\t'.join([get(field) for field in FIELD_NAMES])
    @classmethod
    def from_conll(this_class, text):
        """Construct a Token from a line in CoNLL-X format."""
        fields = text.split('\t')
        fields[0] = int(fields[0]) # index
        fields[6] = int(fields[6]) # head index
        if fields[5] != '_': # feats
            fields[5] = tuple(fields[5].split('|'))
        fields = [value if value != '_' else None for value in fields]
        return this_class(**dict(zip(FIELD_NAMES, fields)))

class Sentence(list):
    """Sequence of Token objects."""
    def as_conll(self):
        """Represent this Sentence as a string in CoNLL-X format."""
        return '\n'.join(token.as_conll() for token in self)
    def as_asciitree(self, str_func=None):
        """Represent this Sentence as an ASCII tree string. Requires
        the asciitree package. A default token stringifier is provided
        but for custom formatting, specify a str_func which should take
        a single Token and return a string."""
        import asciitree
        from collections import defaultdict
        children = defaultdict(list)
        # since erased nodes may be missing, multiple tokens may have same
        # index (CCprocessed), etc.
        token_to_index = {}
        roots = []
        for token in self:
            children[token.head].append(token)
            token_to_index[token] = token.index
            if token.head == 0:
                roots.append(token)
        assert roots, "Couldn't find root Token(s)"

        if len(roots) > 1:
            # multiple roots so we make a fake one to be their parent
            root = Token(0, 'ROOT', 'ROOT-LEMMA', 'ROOT-CPOS', 'ROOT-POS',
                         None, None, 'ROOT-DEPREL', None, None)
            token_to_index[root] = 0
            children[0] = roots
        else:
            root = roots[0]

        def child_func(token):
            index = token_to_index[token]
            return children[index]
        if not str_func:
            def str_func(token):
                return ' %s [%s]' % (token.form, token.deprel)

        return asciitree.draw_tree(root, child_func, str_func)
    def as_dotgraph(self, digraph_kwargs=None, id_prefix=None,
                    node_formatter=None, edge_formatter=None):
        """Returns this sentence as a graphviz.Digraph. Requires the
        graphviz Python package and graphviz itself. There are several
        ways to customize. Graph level keyword arguments can be passed
        as a dictionary to digraph_kwargs. If you're viewing multiple
        Sentences in the same graph, you'll need to set a unique prefix
        string in id_prefix. Lastly, you can change the formatting of
        nodes and edges with node_formatter and edge_formatter. Both
        take a single Token as an argument (for edge_formatter, the
        Token represents the child token) and return a dictionary of
        keyword arguments which are passed to the node and edge creation
        functions in graphviz. The node_formatter will also be called
        with None as its token when adding the root."""
        digraph_kwargs = digraph_kwargs or {}
        id_prefix = id_prefix or ''

        node_formatter = node_formatter or (lambda token: {})
        edge_formatter = edge_formatter or (lambda token: {})

        import graphviz
        graph = graphviz.Digraph(**digraph_kwargs)
        # add root node
        graph.node(id_prefix + '0', 'root', **node_formatter(None))

        # add remaining nodes and edges
        already_added = set()
        for token in self:
            token_id = id_prefix + str(token.index)
            parent_id = id_prefix + str(token.head)
            if token_id not in already_added:
                graph.node(token_id, token.form, **node_formatter(token))
            graph.edge(parent_id, token_id, label=token.deprel,
                       **edge_formatter(token))
            already_added.add(token_id)
        return graph

    @classmethod
    def from_conll(this_class, stream):
        """Construct a Sentence. stream is an iterable over strings where
        each string is a line in CoNLL-X format. If there are multiple
        sentences in this stream, we only return the first one."""
        stream = iter(stream)
        sentence = this_class()
        for line in stream:
            line = line.strip()
            if line:
                sentence.append(Token.from_conll(line))
            elif sentence:
                return sentence
        return sentence
    @classmethod
    def from_stanford_dependencies(this_class, stream, tree,
                                   include_erased=False, include_punct=True):
        """Construct a Sentence. stream is an iterable over strings
        where each string is a line representing a Stanford Dependency
        as in the output of the command line Stanford Dependency tool:

            deprel(gov-index, dep-depindex)

        The corresponding Penn Treebank formatted tree must be provided
        as well."""
        stream = iter(stream)
        sentence = this_class()
        covered_indices = set()
        tags_and_words = ptb_tags_and_words_re.findall(tree)
        for byte_line in stream:
            line = byte_line.decode("UTF-8")
            if not line.strip():
                if sentence:
                    # empty line means the sentence is over
                    break
                else:
                    continue
            matches = deps_re.findall(line)
            assert len(matches) == 1
            deprel, gov_form, head, form, index = matches[0]
            index = int(index)
            tag, word = tags_and_words[index - 1]
            assert form == word
            covered_indices.add(index)

            if not include_punct and deprel == 'punct':
                continue
            token = Token(index, form, None, tag, tag, None, int(head),
                          deprel, None, None)
            sentence.append(token)

        if include_erased:
            # look through words in the tree to see if any of them
            # were erased
            for index, (tag, word) in enumerate(tags_and_words, 1):
                if index in covered_indices:
                    continue
                token = Token(index, word, None, tag, tag, None, 0,
                              'erased', None, None)
                sentence.append(token)

        sentence.sort()
        return sentence

class Corpus(list):
    """Sequence of Sentence objects."""
    def as_conll(self):
        """Represent the entire corpus as a string in CoNLL-X format."""
        return '\n'.join(sentence.as_conll() for sentence in self)
    @classmethod
    def from_conll(this_class, stream):
        """Construct a Corpus. stream is an iterable over strings where
        each string is a line in CoNLL-X format."""
        stream = iter(stream)
        corpus = this_class()
        while 1:
            # read until we get an empty sentence
            sentence = Sentence.from_conll(stream)
            if sentence:
                corpus.append(sentence)
            else:
                break
        return corpus
    @classmethod
    def from_stanford_dependencies(this_class, stream, trees,
                                   include_erased=False, include_punct=True):
        """Construct a Corpus. stream is an iterable over strings where
        each string is a line representing a Stanford Dependency as in
        the output of the command line Stanford Dependency tool:

            deprel(gov-index, dep-depindex)

        Sentences are separated by blank lines. A corresponding list of
        Penn Treebank formatted trees must be provided as well."""
        stream = iter(stream)
        corpus = this_class()
        for tree in trees:
            sentence = Sentence.from_stanford_dependencies(stream,
                                                           tree,
                                                           include_erased,
                                                           include_punct)
            corpus.append(sentence)
        return corpus
